import datetime
import json

import pytest
from mock import MagicMock, call

import mock
import os.path
import inspect

from tap_mambu.helpers.generator_processor_pairs import get_generator_processor_for_stream, \
    get_available_streams
from tap_mambu.tap_generators.child_generator import ChildGenerator
from tap_mambu.tap_processors.child_processor import ChildProcessor
from tap_mambu.tap_processors.deduplication_processor import DeduplicationProcessor
from tap_mambu.helpers import convert

from ..constants import config_json
from ..helpers import GeneratorMock, IsInstanceMatcher

FIXTURES_PATH = f"{os.path.dirname(os.path.abspath(inspect.stack()[0][1]))}/Fixtures"


@mock.patch("tap_mambu.tap_processors.processor.TapProcessor.write_schema")
@mock.patch("tap_mambu.tap_processors.processor.TapProcessor.write_bookmark")
def test_tap_processor_deduplication(mock_write_bookmark,
                                     mock_write_schema,  # Mock write_schema so we don't pollute the output
                                     capsys):
    from tap_mambu import discover
    catalog = discover()

    expected_output = [
        {"encoded_key": "1", "last_modified_date": '2022-01-01T01:00:00.000000Z'},
        {"encoded_key": "2", "last_modified_date": '2022-01-02T00:00:00.000000Z'},
        {"encoded_key": "3", "last_modified_date": '2022-01-03T00:00:00.000000Z'},
        {"encoded_key": "4", "last_modified_date": '2022-01-04T04:00:00.000000Z'},
        {"encoded_key": "5", "last_modified_date": '2022-01-05T05:00:00.000000Z'},
        {"encoded_key": "6", "last_modified_date": '2022-01-06T00:00:00.000000Z'}
    ]
    generators_data = [
        [{"encoded_key": "1", "last_modified_date": '2022-01-01T03:00:00+03:00'},
         {"encoded_key": "4", "last_modified_date": '2022-01-04T07:00:00+03:00'},
         {"encoded_key": "5", "last_modified_date": '2022-01-05T03:00:00+03:00'}],
        [{"encoded_key": "2", "last_modified_date": '2022-01-02T03:00:00+03:00'},
         {"encoded_key": "3", "last_modified_date": '2022-01-03T03:00:00+03:00'},
         {"encoded_key": "4", "last_modified_date": '2022-01-04T03:00:00+03:00'}],
        [{"encoded_key": "1", "last_modified_date": '2022-01-01T04:00:00+03:00'},
         {"encoded_key": "5", "last_modified_date": '2022-01-05T08:00:00+03:00'},
         {"encoded_key": "6", "last_modified_date": '2022-01-06T03:00:00+03:00'}],
    ]
    generators = [GeneratorMock(list(generator_data)) for generator_data in generators_data]
    for generator in generators:
        generator.time_extracted = 0

    client_mock = MagicMock()
    processor = DeduplicationProcessor(catalog=catalog,
                                       stream_name="loan_accounts",
                                       client=client_mock,
                                       config=config_json,
                                       state={'currently_syncing': 'loan_accounts'},
                                       sub_type="self",
                                       generators=generators)
    processor.process_streams_from_generators()
    captured = capsys.readouterr()
    stdout_list = [json.loads(line) for line in captured.out.split("\n") if line]
    # noinspection PyTypeChecker
    assert stdout_list == [
        {"type": "RECORD", "stream": "loan_accounts", "record": record} for record in expected_output
    ], "Output should contain mocked records"


@mock.patch("tap_mambu.tap_processors.multithreaded_parent_processor.get_selected_streams")
@mock.patch("tap_mambu.tap_processors.processor.TapProcessor.write_schema")
@mock.patch("tap_mambu.tap_processors.processor.TapProcessor.write_bookmark")
@mock.patch("tap_mambu.sync.sync_endpoint")
def test_tap_processor_process_child_records(mock_sync_endpoint_refactor,
                                             mock_write_bookmark,
                                             mock_write_schema,  # Mock write_schema so we don't pollute the output
                                             mock_get_selected_streams,
                                             capsys):
    from tap_mambu import discover
    from tap_mambu.tap_processors.multithreaded_parent_processor import MultithreadedParentProcessor
    fake_children_record_count = 4
    mock_get_selected_streams.return_value = ["child_1", "child_2"]
    mock_sync_endpoint_refactor.return_value = fake_children_record_count
    catalog = discover()

    generator_data = [
        {
            "encoded_key": "12345678901234567890123456789012",
            "last_modified_date": "2022-01-01T00:00:00.000000Z",
            "id": "2"
        }, {
            "encoded_key": "12345678901234567890123456789013",
            "last_modified_date": "2022-01-01T00:00:00.000000Z",
            "id": "3"
        }, {
            "encoded_key": "12345678901234567890123456789014",
            "last_modified_date": "2022-01-01T00:00:00.000000Z",
            "id": "4"
        }, {
            "encoded_key": "12345678901234567890123456789015",
            "last_modified_date": "2022-01-01T00:00:00.000000Z",
            "id": "5"
        }
    ]
    generator = GeneratorMock(list(generator_data))
    generator.time_extracted = 0

    client_mock = MagicMock()
    processor = MultithreadedParentProcessor(catalog=catalog,
                                             stream_name="loan_accounts",
                                             client=client_mock,
                                             config=config_json,
                                             state={'currently_syncing': 'loan_accounts'},
                                             sub_type="self",
                                             generators=[generator])
    processor.endpoint_child_streams = ["child_1", "child_2"]
    actual_records_count = processor.process_streams_from_generators()
    # sync_endpoint_refactor called for every record (len(generator_data)) for every child_stream + once for parent
    assert actual_records_count == \
           len(generator_data) * (fake_children_record_count * len(processor.endpoint_child_streams) + 1), \
           "Record count mismatch when adding child records"

    mock_sync_endpoint_refactor.assert_called_with(client=client_mock,
                                                   catalog=processor.catalog,
                                                   state={'currently_syncing': 'loan_accounts'},
                                                   stream_name=processor.endpoint_child_streams[-1],
                                                   sub_type="self",
                                                   config=config_json,
                                                   parent_id="5")

    captured = capsys.readouterr()
    stdout_list = [json.loads(line) for line in captured.out.split("\n") if line]
    # noinspection PyTypeChecker
    assert stdout_list == [
        {"type": "RECORD", "stream": "loan_accounts", "record": record} for record in generator_data
    ], "Output should contain mocked records"


@mock.patch("tap_mambu.helpers.write_state")
def test_bookmarks(mock_write_state):
    from tap_mambu import discover
    from tap_mambu.tap_processors.processor import TapProcessor

    catalog = discover()
    client_mock = MagicMock()
    processor = TapProcessor(catalog=catalog,
                             stream_name="loan_accounts",
                             client=client_mock,
                             config=config_json,
                             state={'currently_syncing': 'loan_accounts'},
                             sub_type="self",
                             generators=[GeneratorMock([])])

    processor.write_bookmark()

    expected_state = {'currently_syncing': 'loan_accounts', 'bookmarks': {
        'loan_accounts': '2021-06-01T00:00:00Z'}}
    mock_write_state.assert_called_once_with(expected_state)


@mock.patch("tap_mambu.tap_processors.processor.write_schema")
def test_write_schema(mock_write_schema):
    from singer.catalog import Catalog
    from tap_mambu.tap_processors.processor import TapProcessor

    catalog = Catalog.load(f"{FIXTURES_PATH}/processor_catalog.json")
    client_mock = MagicMock()
    processor = TapProcessor(catalog=catalog,
                             stream_name="loan_accounts",
                             client=client_mock,
                             config=config_json,
                             state={'currently_syncing': 'loan_accounts'},
                             sub_type="self",
                             generators=[GeneratorMock([])])
    processor.write_schema()

    schema = None
    stream_key_properties = None
    with open(f"{FIXTURES_PATH}/processor_catalog.json", "r") as fd:
        schema_json = json.loads(fd.read())
        for stream in schema_json["streams"]:
            if stream["tap_stream_id"] == "loan_accounts":
                schema = stream["schema"]
                stream_key_properties = stream["key_properties"]
                break
    assert schema is not None
    mock_write_schema.assert_called_with("loan_accounts", schema, stream_key_properties)


@mock.patch("tap_mambu.tap_processors.processor.write_record")
@mock.patch("tap_mambu.tap_processors.processor.write_schema")
def test_write_exceptions(mock_write_schema, mock_write_record):
    from tap_mambu import discover
    from tap_mambu.tap_processors.processor import TapProcessor
    catalog = discover()

    mock_write_record.side_effect = [None, OSError("Mock Record Exception")]
    mock_write_schema.side_effect = [None, OSError("Mock Schema Exception")]

    client_mock = MagicMock()
    processor = TapProcessor(catalog=catalog,
                             stream_name="loan_accounts",
                             client=client_mock,
                             config=config_json,
                             state={'currently_syncing': 'loan_accounts'},
                             sub_type="self",
                             generators=[GeneratorMock([
                                 {"id": "1", "last_modified_date": "2022-01-01T00:00:00+03:00"},
                                 {"id": "2", "last_modified_date": "2022-01-01T01:00:00+03:00"},
                                 {"id": "3", "last_modified_date": "2022-01-01T02:00:00+03:00"}
                             ])])
    processor.endpoint_deduplication_key = "id"

    with pytest.raises(OSError) as err:
        processor.process_streams_from_generators()
    assert err.value.args[0] == "Mock Record Exception"

    mock_write_record.assert_has_calls([
        call("loan_accounts", IsInstanceMatcher(dict), time_extracted=IsInstanceMatcher(datetime.datetime)),
        call("loan_accounts", IsInstanceMatcher(dict), time_extracted=IsInstanceMatcher(datetime.datetime))
    ])

    with pytest.raises(OSError) as err:
        processor.process_streams_from_generators()
    assert err.value.args[0] == "Mock Schema Exception"

    mock_write_schema.assert_has_calls([
        call("loan_accounts", IsInstanceMatcher(dict), IsInstanceMatcher(list)),
        call("loan_accounts", IsInstanceMatcher(dict), IsInstanceMatcher(list))
    ])


def test_catalog_automatic_fields():
    from tap_mambu import discover

    client_mock = MagicMock()
    client_mock.page_size = 5
    client_mock.request = MagicMock()
    catalog = discover()

    for stream in get_available_streams():
        catalog_stream = catalog.get_stream(stream)
        generator_classes, processor_class = get_generator_processor_for_stream(stream)

        automatic_fields = [
            mdata["breadcrumb"][1]
            for mdata in catalog_stream.metadata
            if mdata["breadcrumb"] and mdata["metadata"]["inclusion"] == "automatic"
        ]

        generator = None
        for generator_class in generator_classes:
            generator = generator_class(stream_name=stream,
                                        client=client_mock,
                                        config=config_json,
                                        state={"currently_syncing": stream},
                                        sub_type="self",
                                        **({"parent_id": "0"} if issubclass(generator_class, ChildGenerator) else {}))
            if generator.endpoint_bookmark_field != "":
                # Those streams do not respect the camelCase convention for field names
                if generator.stream_name not in ["audit_trail"]:
                    assert "_" not in generator.endpoint_bookmark_field,\
                        f"Generator bookmark field for '{stream}' stream should be in camelCase!"
                assert convert(generator.endpoint_bookmark_field) in automatic_fields,\
                    f"Generator bookmark field for '{stream}' stream should be set to automatic in catalog!"

        # We need to mock iter, so we don't trigger preloading data
        # (as the processor would when instantiating the generator)
        with mock.patch.object(generator, "__iter__") as mock_iter:
            mock_iter.return_value = generator
            processor = processor_class(catalog=catalog,
                                        stream_name="loan_accounts",
                                        client=client_mock,
                                        config=config_json,
                                        state={'currently_syncing': 'loan_accounts'},
                                        sub_type="self",
                                        generators=[generator],
                                        **({"parent_id": "0"} if issubclass(processor_class, ChildProcessor) else {}))

            if isinstance(processor, DeduplicationProcessor):
                assert all([char.islower() for char in processor.endpoint_deduplication_key if char != "_"]),\
                            f"Processor deduplication key for '{stream}' stream should be in snake_case!"
                assert processor.endpoint_deduplication_key in automatic_fields,\
                            f"Processor deduplication key for '{stream}' stream should be set to automatic in catalog!"



from . import setup_processor_base_test


def test_loan_repayments_processor_endpoint_config_init():
    processor = setup_processor_base_test("loan_repayments")

    assert processor.endpoint_parent == 'loan_accounts'
    assert processor.endpoint_parent_id == 'TEST'



from .processor import TapProcessor
from singer.utils import strptime_to_utc
from ..helpers import convert, write_bookmark, get_bookmark


class AuditTrailProcessor(TapProcessor):
    def __init__(self, *args, **kwargs):
        super(AuditTrailProcessor, self).__init__(*args, **kwargs)
        # This will be used to signal the generator how many records to skip (when more than one record has same date)
        self.bookmark_offset = 1

    def _init_endpoint_config(self):
        # There are no unique fields in this stream
        self.endpoint_deduplication_key = None

    def _init_bookmarks(self):
        super(AuditTrailProcessor, self)._init_bookmarks()
        self.last_bookmark_value = get_bookmark(self.state, self.stream_name, self.sub_type, self.start_date)
        if type(self.last_bookmark_value) is list:
            if len(self.last_bookmark_value) != 2:
                raise ValueError("Cannot parse audit trail bookmark from list because we expect 2 values!")
            self.last_bookmark_value = self.last_bookmark_value[0]
        self.max_bookmark_value = self.last_bookmark_value

    def _update_bookmark(self, transformed_record, bookmark_field):
        bookmark_field = convert(bookmark_field)
        if bookmark_field and (bookmark_field in transformed_record):
            bookmark_dttm = strptime_to_utc(transformed_record[bookmark_field])
            max_bookmark_value_dttm = strptime_to_utc(self.max_bookmark_value)
            if bookmark_dttm > max_bookmark_value_dttm:
                self.max_bookmark_value = bookmark_dttm.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
                # Reset identical record count
                self.bookmark_offset = 0

            max_bookmark_value_dttm = strptime_to_utc(self.max_bookmark_value)
            if max_bookmark_value_dttm == bookmark_dttm:
                # Increment identical record count
                self.bookmark_offset += 1

        for generator in self.generators:
            # Bad practice, but necessary as we need to change extraction parameters while extracting
            generator.static_params["occurred_at[gte]"] = self.max_bookmark_value

    def write_bookmark(self):
        if all([generator.endpoint_bookmark_field for generator in self.generators]):
            write_bookmark(self.state,
                           self.stream_name,
                           self.sub_type,
                           [self.max_bookmark_value, self.bookmark_offset])
            # Saving bookmark as a list, containing the bookmark value, as well as the number of records extracted on
            # the same date as the bookmark so we skip them at the next extraction



from singer import metadata
from singer.catalog import Catalog, CatalogEntry, Schema
from tap_mambu.helpers.schema import get_schemas, STREAMS


def discover():
    schemas, field_metadata = get_schemas()
    catalog = Catalog([])

    for stream_name, schema_dict in schemas.items():
        schema = Schema.from_dict(schema_dict)
        mdata = metadata.to_map(field_metadata[stream_name])

        stream = STREAMS[stream_name]
        if stream.get('replication_method') == 'INCREMENTAL':
            for field_name in stream.get('replication_keys'):
                metadata.write(mdata, ('properties', field_name), 'inclusion', 'automatic')

        catalog.streams.append(CatalogEntry(
            stream=stream_name,
            tap_stream_id=stream_name,
            key_properties=stream['key_properties'],
            schema=schema,
            metadata=metadata.to_list(mdata)
        ))

    return catalog



from . import setup_processor_base_test


def test_deposit_cards_processor_endpoint_config_init():
    processor = setup_processor_base_test("cards")

    assert processor.endpoint_id_field == 'reference_token'
    assert processor.endpoint_parent == 'deposit'
    assert processor.endpoint_parent_id == 'TEST'



import inspect
import os

import mock


FIXTURES_PATH = f"{os.path.dirname(os.path.abspath(inspect.stack()[0][1]))}/Fixtures"


def test_get_selected_streams():
    from singer.catalog import Catalog
    from tap_mambu.helpers import get_selected_streams
    catalog = Catalog.load(f"{FIXTURES_PATH}/catalog.json")
    selected_streams = get_selected_streams(catalog)
    expected_streams = ["loan_accounts", "loan_repayments", "audit_trail"]
    assert len(selected_streams) == len(expected_streams) and set(selected_streams) == set(expected_streams)


def test_get_bookmark():
    from tap_mambu.helpers import get_bookmark
    state = {"currently_syncing": "loan_accounts",
             "bookmarks": {"loan_accounts": "2021-10-01T00:00:00Z"}}
    assert get_bookmark(state, "loan_accounts", "self", "2021-06-01T00:00:00Z") == "2021-10-01T00:00:00Z"

    state = {"currently_syncing": "loan_accounts",
             "bookmarks": {"loan_accounts": "2021-10-01T00:00:00Z"}}
    assert get_bookmark(state, "deposit_accounts", "self", "2021-06-01T00:00:00Z") == "2021-06-01T00:00:00Z"

    state = {"currently_syncing": "loan_accounts"}
    assert get_bookmark(state, "loan_accounts", "self", "2021-06-01T00:00:00Z") == "2021-06-01T00:00:00Z"


def test_get_bookmark_sub_type():
    from tap_mambu.helpers import get_bookmark
    state = {"currently_syncing": "loan_accounts",
             "bookmarks": {"loan_accounts": {
                 "1": "2021-10-01T00:00:00Z"
             }}}
    assert get_bookmark(state, "loan_accounts", "1", "2021-06-01T00:00:00Z") == "2021-10-01T00:00:00Z"


@mock.patch("tap_mambu.helpers.write_state")
def test_write_bookmark(mock_write_state):
    from tap_mambu.helpers import write_bookmark

    state = {"currently_syncing": "loan_accounts"}
    write_bookmark(state, "loan_accounts", "self", "2021-10-01T00:00:00Z")

    mock_write_state.assert_called_with({"currently_syncing": "loan_accounts",
                                         "bookmarks": {"loan_accounts": "2021-10-01T00:00:00Z"}})

    state = {"currently_syncing": "loan_accounts",
             "bookmarks": {"loan_accounts": "2021-08-01T00:00:00Z"}}
    write_bookmark(state, "loan_accounts", "self", "2021-11-01T00:00:00Z")

    mock_write_state.assert_called_with({"currently_syncing": "loan_accounts",
                                         "bookmarks": {"loan_accounts": "2021-11-01T00:00:00Z"}})


@mock.patch("tap_mambu.helpers.write_state")
def test_write_bookmark_sub_type(mock_write_state):
    from tap_mambu.helpers import write_bookmark

    state = {"currently_syncing": "loan_accounts"}
    write_bookmark(state, "loan_accounts", "1", "2021-10-01T00:00:00Z")

    mock_write_state.assert_called_with({"currently_syncing": "loan_accounts",
                                         "bookmarks": {"loan_accounts": {
                                             "1": "2021-10-01T00:00:00Z"
                                         }}})

    state = {"currently_syncing": "loan_accounts",
             "bookmarks": {"loan_accounts": {
                 "1": "2021-08-01T00:00:00Z"
             }}}
    write_bookmark(state, "loan_accounts", "1", "2021-11-01T00:00:00Z")

    mock_write_state.assert_called_with({"currently_syncing": "loan_accounts",
                                         "bookmarks": {"loan_accounts": {
                                             "1": "2021-11-01T00:00:00Z"
                                         }}})



import json


class HashableDict(dict):

    @staticmethod
    def _recur_hash(value):
        if type(value) in [dict, HashableDict]:
            return HashableDict(value).__key()
        if type(value) == list:
            return tuple(sorted(map(HashableDict._recur_hash, value)))
        return value

    def __key(self):
        data = [(key, self._recur_hash(value)) for key, value in self.items()]
        return tuple(sorted(data, key=lambda record: record[0]))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if isinstance(other, HashableDict):
            return self.__key() == other.__key()
        return NotImplemented



from abc import ABC
from concurrent import futures

from singer import write_record, metadata, write_schema, get_logger, metrics, Transformer
from singer.utils import strptime_to_utc, now as singer_now

from ..helpers import convert, get_bookmark, write_bookmark, transform_datetime
from ..helpers.exceptions import NoDeduplicationCapabilityException
from ..helpers.perf_metrics import PerformanceMetrics
# from ..helpers.transformer import Transformer

LOGGER = get_logger()


class TapProcessor(ABC):
    def __init__(self, catalog, stream_name, client, config, state, sub_type, generators):
        self.start_time = singer_now()
        self.generators = generators
        self.generator_values = dict()
        for generator in self.generators:
            self.generator_values[generator.__iter__()] = None
        self.catalog = catalog
        self.stream_name = stream_name
        self.client = client
        self.config = config
        self.state = state
        self.sub_type = sub_type
        self.stream = self.catalog.get_stream(stream_name)
        self.schema = self.stream.schema.to_dict()
        self.stream_metadata = metadata.to_map(self.stream.metadata)
        self.futures = list()
        self._init_config()
        self._init_endpoint_config()
        self._init_bookmarks()
        LOGGER.info(f'(processor) Stream {self.stream_name} - bookmark value used: {self.last_bookmark_value}')

    def _init_config(self):
        self.start_date = self.config.get('start_date')

    def _init_endpoint_config(self):
        self.endpoint_child_streams = []
        self.endpoint_id_field = "id"

        if len(self.generators) > 1:
            raise NoDeduplicationCapabilityException("In order to merge streams in the processor, "
                                                     "you need to use the deduplication processor")

    def _init_bookmarks(self):
        self.last_bookmark_value = get_bookmark(self.state, self.stream_name, self.sub_type, self.start_date)
        self.max_bookmark_value = self.last_bookmark_value

    def write_schema(self):
        stream = self.catalog.get_stream(self.stream_name)
        schema = stream.schema.to_dict()
        try:
            write_schema(self.stream_name, schema, stream.key_properties)
        except OSError as err:
            LOGGER.info(f'OS Error writing schema for: {self.stream_name}')
            raise err

    def process_records(self):
        record_count = 0
        with metrics.record_counter(self.stream_name) as counter:
            for record in self.generators[0]:
                # Process the record
                with PerformanceMetrics(metric_name="processor"):
                    is_processed = self.process_record(record, self.generators[0].time_extracted,
                                                       self.generators[0].endpoint_bookmark_field)
                if is_processed:
                    record_count += 1
                    self._process_child_records_multithreaded(record)
                    counter.increment()

        for future in futures.as_completed(self.futures):
            record_count += future.result()
        return record_count

    def process_streams_from_generators(self):
        self.write_schema()

        record_count = self.process_records()
        self.write_bookmark()
        return record_count

    def _process_child_records_multithreaded(self, record):
        pass

    def _update_bookmark(self, transformed_record, bookmark_field):
        bookmark_field = convert(bookmark_field)
        if bookmark_field and (bookmark_field in transformed_record):
            bookmark_dttm = strptime_to_utc(transformed_record[bookmark_field])
            max_bookmark_value_dttm = strptime_to_utc(self.max_bookmark_value)
            if bookmark_dttm > max_bookmark_value_dttm:
                self.max_bookmark_value = min(bookmark_dttm,
                                              self.start_time).strftime("%Y-%m-%dT%H:%M:%S.%fZ")

    def _is_record_past_bookmark(self, transformed_record, bookmark_field):
        bookmark_field = convert(bookmark_field)

        # If stream doesn't have a bookmark field or the record doesn't contain the stream's bookmark field
        if not bookmark_field or (bookmark_field not in transformed_record):
            return True

        # Keep only records whose bookmark is after the last_datetime
        if transform_datetime(transformed_record[bookmark_field]) >= transform_datetime(self.last_bookmark_value):
            return True
        return False

    def process_record(self, record, time_extracted, bookmark_field):
        with Transformer() as transformer:
            transformed_record = transformer.transform(record,
                                                       self.schema,
                                                       self.stream_metadata)

        self._update_bookmark(transformed_record, bookmark_field)
        if self._is_record_past_bookmark(transformed_record, bookmark_field):
            try:
                write_record(self.stream_name,
                             transformed_record,
                             time_extracted=time_extracted)
            except OSError as err:
                LOGGER.info(f'OS Error writing record for: {self.stream_name}')
                LOGGER.info(f'record: {transformed_record}')
                raise err

            return True
        return False

    def write_bookmark(self):
        if all([generator.endpoint_bookmark_field for generator in self.generators]):
            write_bookmark(self.state,
                           self.stream_name,
                           self.sub_type,
                           self.max_bookmark_value)



from tap_mambu.tap_generators.generator import TapGenerator


class NoPaginationGenerator(TapGenerator):
    def _init_params(self):
        self.time_extracted = None

    def prepare_batch(self):
        self.params = {}

    def __next__(self):
        if not self.buffer:
            raise StopIteration()
        return self.buffer.pop(0)



from ..tap_generators.activities_generator import ActivitiesGenerator
from ..tap_generators.audit_trail_generator import AuditTrailGenerator
from ..tap_generators.branches_generator import BranchesGenerator
from ..tap_generators.centres_generator import CentresGenerator
from ..tap_generators.clients_generator import ClientsGenerator
from ..tap_generators.communications_generator import CommunicationsGenerator
from ..tap_generators.credit_arrangements_generator import CreditArrangementsGenerator
from ..tap_generators.custom_field_sets_generator import CustomFieldSetsGenerator
from ..tap_generators.deposit_accounts_generator import DepositAccountsGenerator
from ..tap_generators.deposit_cards_generator import DepositCardsGenerator
from ..tap_generators.deposit_products_generator import DepositProductsGenerator
from ..tap_generators.gl_accounts_generator import GlAccountsGenerator
from ..tap_generators.gl_journal_entries_generator import GlJournalEntriesGenerator
from ..tap_generators.installments_generator import InstallmentsGenerator
from ..tap_generators.groups_generator import GroupsGenerator
from ..tap_generators.index_rate_sources_generator import IndexRateSourcesGenerator
from ..tap_generators.interest_accrual_breakdown_generator import InterestAccrualBreakdownGenerator
from ..tap_generators.deposit_transactions_generator import DepositTransactionsGenerator
from ..tap_generators.loan_accounts_generator import LoanAccountsADGenerator, LoanAccountsLMGenerator
from ..tap_generators.loan_products_generator import LoanProductsGenerator
from ..tap_generators.loan_repayments_generator import LoanRepaymentsGenerator
from ..tap_generators.loan_transactions_generator import LoanTransactionsGenerator
from ..tap_generators.tasks_generator import TasksGenerator
from ..tap_generators.users_generator import UsersGenerator
from ..tap_processors.audit_trail_processor import AuditTrailProcessor
from ..tap_processors.deposit_accounts_processor import DepositAccountsProcessor
from ..tap_processors.deposit_cards_processor import DepositCardsProcessor
from ..tap_processors.loan_accounts_processor import LoanAccountsProcessor
from ..tap_processors.loan_repayments_processor import LoanRepaymentsProcessor
from ..tap_processors.processor import TapProcessor


def get_generator_processor_pairs():
    return {
        "activities": ((ActivitiesGenerator,), TapProcessor),
        "audit_trail": ((AuditTrailGenerator,), AuditTrailProcessor),
        "branches": ((BranchesGenerator,), TapProcessor),
        "cards": ((DepositCardsGenerator,), DepositCardsProcessor),
        "centres": ((CentresGenerator,), TapProcessor),
        "clients": ((ClientsGenerator,), TapProcessor),
        "communications": ((CommunicationsGenerator,), TapProcessor),
        "credit_arrangements": ((CreditArrangementsGenerator,), TapProcessor),
        "custom_field_sets": ((CustomFieldSetsGenerator,), TapProcessor),
        "deposit_accounts": ((DepositAccountsGenerator,), DepositAccountsProcessor),
        "deposit_products": ((DepositProductsGenerator,), TapProcessor),
        "gl_accounts": ((GlAccountsGenerator,), TapProcessor),
        "gl_journal_entries": ((GlJournalEntriesGenerator,), TapProcessor),
        "groups": ((GroupsGenerator,), TapProcessor),
        "index_rate_sources": ((IndexRateSourcesGenerator,), TapProcessor),
        "deposit_transactions": ((DepositTransactionsGenerator,), TapProcessor),
        "installments": ((InstallmentsGenerator,), TapProcessor),
        "interest_accrual_breakdown": ((InterestAccrualBreakdownGenerator,), TapProcessor),
        "loan_accounts": ((LoanAccountsLMGenerator, LoanAccountsADGenerator), LoanAccountsProcessor),
        "loan_products": ((LoanProductsGenerator,), TapProcessor),
        "loan_repayments": ((LoanRepaymentsGenerator,), LoanRepaymentsProcessor),
        "loan_transactions": ((LoanTransactionsGenerator,), TapProcessor),
        "tasks": ((TasksGenerator,), TapProcessor),
        "users": ((UsersGenerator,), TapProcessor),
    }


def get_available_streams():
    return list(get_generator_processor_pairs().keys())


def get_generator_processor_for_stream(stream: str):
    stream_generator_processor_dict = get_generator_processor_pairs()
    return stream_generator_processor_dict[stream]


def get_stream_subtypes(stream: str):
    if stream in ["gl_accounts"]:
        return ["ASSET", "LIABILITY", "EQUITY", "INCOME", "EXPENSE"]
    return ["self"]



from singer.utils import load_json
from tap_mambu.client import MambuClient
from schema_fetcher.constants import CLIENT_CONFIG_FILE_PATH, TAP_TYPES_AND_FORMATS, TAP_FORMATS


def get_mambu_client():
    config = load_json(CLIENT_CONFIG_FILE_PATH)
    return MambuClient(username=config.get('username'),
                       password=config.get('password'),
                       apikey='',
                       subdomain=config.get('subdomain'),
                       apikey_audit='',
                       page_size='',
                       user_agent='')


def get_data_type_and_format(field_type, field_format=None, obj_type_nullable=True):
    tap_field_type = TAP_TYPES_AND_FORMATS[field_type]
    if not obj_type_nullable and isinstance(tap_field_type, list):
        tap_field_type = tap_field_type[1]

    type_and_format = {'type': tap_field_type}
    if field_format and TAP_FORMATS[field_format]:
        type_and_format['format'] = TAP_FORMATS[field_format]
    if field_type == 'number':
        type_and_format['format'] = 'singer.decimal'
    return type_and_format



import mock
from singer import utils
from . import setup_generator_base_test


def test_activities_generator_endpoint_config_init():
    generators = setup_generator_base_test("activities")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.endpoint_path == 'activities'
    assert generator.endpoint_api_method == "GET"
    assert generator.endpoint_api_version == "v1"
    assert generator.endpoint_params == {
        "from": "2021-06-01",
        "to": utils.now().strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:10],
        "detailsLevel": "FULL",
        "paginationDetails": "OFF"
    }
    assert generator.endpoint_bookmark_field == "timestamp"


def test_activities_generator_dict_unpacking():
    generators = setup_generator_base_test("activities")

    assert 1 == len(generators)

    generator = generators[0]

    generator.client.request.side_effect = [[
        {"client": "N/A", "activity": {"id": nr + page*4}}
        for nr in range(5)] for page in range(3)] + [[] for _ in range(1000)] + [[] * 1000]

    for record in generator:
        assert "id" in record, "Record in activity generator were not unpacked " \
                               "according to spec! (see fetch_batch)"
        assert "client" in record, "Record in activity generator were not unpacked " \
                                   "according to spec! (see fetch_batch)"
        assert "activity" not in record, "Record in activity generator were not unpacked " \
                                         "according to spec! (see fetch_batch)"



from . import setup_generator_base_test


def test_custom_field_sets_generator_endpoint_config_init():
    generators = setup_generator_base_test("custom_field_sets")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.endpoint_path == "customfieldsets"
    assert generator.endpoint_api_method == "GET"



import math
import time


class PerformanceMetrics:
    _metrics_start_time = time.monotonic()
    _metrics = dict(generator=list(),
                    processor=list(),
                    generator_wait=list(),
                    processor_wait=list())
    _generator_batch_size = 500

    def __init__(self, metric_name):
        self.start_time = None
        if metric_name not in self._metrics:
            raise ValueError("One argument must be True, but only one!")
        self._metric_name = metric_name

    def __enter__(self):
        self.start_time = time.monotonic()

    def __exit__(self, exc_type, exc_val, exc_tb):
        metric = (self.start_time, time.monotonic())
        self._metrics[self._metric_name].append(metric)

    @classmethod
    def reset_metrics(cls):
        cls._metrics_start_time = time.monotonic()
        cls._metrics = dict(generator=list(),
                            processor=list(),
                            generator_wait=list(),
                            processor_wait=list())

    @classmethod
    def set_generator_batch_size(cls, batch_size):
        if any(cls._metrics.values()):
            raise RuntimeError("You cannot change batch size after measuring metrics!")
        cls._generator_batch_size = batch_size

    @property
    def generator_batch_size(self):
        return self._generator_batch_size

    @staticmethod
    def get_sum(metric):
        if not metric:
            return 0
        return sum([record[1] - record[0] for record in metric])

    @staticmethod
    def get_avg_with_98th(metric):
        if not metric:
            return 0, 0
        values_total = sorted([record[1] - record[0] for record in metric], reverse=True)
        values_98th = values_total[:math.ceil(len(values_total) * 2 / 100)]

        average = sum(values_total) / len(values_total)
        average_98th = sum(values_98th) / len(values_98th)

        return average, average_98th

    @classmethod
    def get_statistics(cls):
        extraction_duration = time.monotonic() - cls._metrics_start_time

        generator_avg, generator_avg_98th = cls.get_avg_with_98th(cls._metrics["generator"])
        processor_avg, processor_avg_98th = cls.get_avg_with_98th(cls._metrics["processor"])
        generator_wait = cls.get_sum(cls._metrics["generator_wait"])
        processor_wait = cls.get_sum(cls._metrics["processor_wait"])

        return dict(generator=generator_avg / cls._generator_batch_size,
                    generator_98th=generator_avg_98th / cls._generator_batch_size,
                    processor=processor_avg, processor_98th=processor_avg_98th,
                    generator_wait=generator_wait,
                    processor_wait=processor_wait,
                    records=len(cls._metrics["processor"]) // extraction_duration,
                    extraction=extraction_duration)



from singer import utils
from . import setup_generator_base_test


def test_installments_generator_endpoint_config_init():
    generators = setup_generator_base_test("installments")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.endpoint_path == 'installments'
    assert generator.endpoint_api_method == "GET"
    assert generator.endpoint_params == {
        "dueFrom": "2021-06-01",
        "dueTo": utils.now().strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:10],
        "detailsLevel": "FULL",
        "paginationDetails": "OFF"
    }
    assert generator.endpoint_bookmark_field == "lastPaidDate"



import inspect
import os
from copy import deepcopy

from unittest.mock import MagicMock

from tap_mambu.helpers import transform_json
from tap_mambu.tap_generators.generator import TapGenerator
from ..constants import config_json

FIXTURES_PATH = f"{os.path.dirname(os.path.abspath(inspect.stack()[0][1]))}/fixtures"


def test_generator_fetch_batch_flow():
    client_mock = MagicMock()
    client_mock.page_size = 5
    client_mock.request = MagicMock()
    client_mock.request.side_effect = [[{"id": nr + page*5} for nr in range(5)] for page in range(3)]
    generator = TapGenerator(stream_name="loan_accounts",
                             client=client_mock,
                             config=config_json,
                             state={"currently_syncing": "loan_accounts"},
                             sub_type="self")
    generator_endpoint_api_method = generator.endpoint_api_method
    generator_endpoint_path = generator.endpoint_path
    generator_endpoint_api_version = generator.endpoint_api_version
    generator_endpoint_api_key_type = generator.endpoint_api_key_type
    generator_endpoint_body = {"sortingCriteria": generator.endpoint_sorting_criteria,
                               "filterCriteria": generator.endpoint_filter_criteria}

    for batch_index in range(3):
        generator.prepare_batch()
        generator_endpoint_querystring = '&'.join([f'{key}={value}' for (key, value) in generator.params.items()])

        assert {"detailsLevel": "FULL", "limit": 5,
                "offset": 0 + batch_index * 5, "paginationDetails": "OFF"} == generator.params
        raw_batch = generator.fetch_batch()
        client_mock.request.assert_called_with(method=generator_endpoint_api_method,
                                               path=generator_endpoint_path,
                                               version=generator_endpoint_api_version,
                                               apikey_type=generator_endpoint_api_key_type,
                                               params=generator_endpoint_querystring,
                                               endpoint="loan_accounts",
                                               json=generator_endpoint_body)

        assert raw_batch == [{"id": nr + batch_index * 5} for nr in range(5)]

        generator.buffer = transform_json(raw_batch, "loan_accounts")
        assert generator.buffer == [{"custom_fields": [], "id": nr + batch_index * 5} for nr in range(5)]

        generator.last_batch_size = len(generator.buffer)

        # Simulate end of batch, so we can fetch a new one
        generator.buffer = []
        generator.offset += generator.limit


def test_generator_iterator_flow():
    client_mock = MagicMock()
    client_mock.page_size = 5
    client_data = [[{"id": nr + page*5} for nr in range(5)] for page in range(3)]
    generator = TapGenerator(stream_name="loan_accounts",
                             client=client_mock,
                             config=config_json,
                             state={"currently_syncing": "loan_accounts"},
                             sub_type="self")
    generator.prepare_batch = MagicMock()
    generator.fetch_batch = MagicMock(side_effect=client_data)

    expected_records = [{"custom_fields": [], "id": record["id"]} for record in deepcopy(sum(client_data, []))]
    record_count = 0
    for record in generator:
        assert record == expected_records[record_count]
        record_count += 1


def test_transform_batch_custom_field():
    from tap_mambu.helpers import transform_json

    assert transform_json([
        {"encodedKey": "123", "creation_date": "asd",
         "_random_field": {"cevaId": "asd"},
         "_random_field_2": [{"a": "b", "c": "d"}],
         "_random_field_3": 0}
    ], "loan_accounts") == [
        {"encoded_key": "123", "creation_date": "asd", "custom_fields": [
            {"field_set_id": "_random_field", "id": "cevaId", "value": "asd"},
            {"field_set_id": "_random_field_2", "id": "a", "value": "b"},
            {"field_set_id": "_random_field_2", "id": "c", "value": "d"}
        ]}
    ]


def test_transform_batch_custom_field_auto_add():
    from tap_mambu.helpers import transform_json

    assert transform_json([
        dict(encodedKey="123", creation_date="asd", _test=2)
    ], "loan_accounts") == [
        dict(encoded_key="123", creation_date="asd", custom_fields=list())
    ]


def test_transform_batch_larger_batch():
    from tap_mambu.helpers import transform_json

    assert transform_json([
        dict(encodedKey="123", creation_date="asd", custom_fields=list()) for _ in range(10)
    ], "loan_accounts") == [
        dict(encoded_key="123", creation_date="asd", custom_fields=list()) for _ in range(10)
    ]



import os

WORKING_DIR_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), '../')

CLIENT_CONFIG_FILE_NAME = 'config.json'
CLIENT_CONFIG_FILE_PATH = os.path.join(WORKING_DIR_PATH, f'{CLIENT_CONFIG_FILE_NAME}')

RESOURCES_LINK = r'swagger/resources/'

OUTPUT_DIR_PATH = os.path.join(os.path.join(WORKING_DIR_PATH, 'tap_mambu'), 'generated_schemas')

TAP_TYPES_AND_FORMATS = {
    'integer': ['null', 'integer'],
    'number': ['null', 'string'],
    'string': ['null', 'string'],
    'boolean': ['null', 'boolean'],
    'object': ['null', 'object'],
    'array': 'array',
}

TAP_FORMATS = {
    'int32': None,
    'int64': None,
    'date': 'date',
    'date-time': 'date-time',
}

CUSTOM_FIELDS_FIELD = {
    "custom_fields": {
        "anyOf": [
            {
                "type": "array",
                "items": {
                    "type": "object",
                    "additionalProperties": False,
                    "properties": {
                        "field_set_id": {
                            "type": [
                                "null",
                                "string"
                            ]
                        },
                        "id": {
                            "type": [
                                "null",
                                "string"
                            ]
                        },
                        "value": {
                            "type": [
                                "null",
                                "string"
                            ]
                        }
                    }
                }
            },
            {
                "type": "null"
            }
        ]
    }}



from . import setup_generator_base_test


def test_repayments_generator_endpoint_config_init():
    generators = setup_generator_base_test("loan_repayments")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.endpoint_api_key_type is None
    assert generator.endpoint_parent_id == "TEST"
    assert generator.endpoint_api_version == "v1"



from .multithreaded_parent_processor import MultithreadedParentProcessor


class DepositAccountsProcessor(MultithreadedParentProcessor):
    def _init_endpoint_config(self):
        super(DepositAccountsProcessor, self)._init_endpoint_config()
        self.endpoint_child_streams = ["cards"]



import datetime
import pytz

from mock import MagicMock, Mock

from tap_mambu.tap_generators.multithreaded_bookmark_generator import MultithreadedOffsetGenerator, \
    MultithreadedBookmarkGenerator, MultithreadedBookmarkDayByDayGenerator


class IsInstanceMatcher:
    def __init__(self, expected_type):
        self.expected_type = expected_type

    def __eq__(self, other):
        return isinstance(other, self.expected_type)


class IsInstanceMockMatcher:
    def __init__(self, expected_type):
        self.expected_type = expected_type

    def __eq__(self, other):
        return isinstance(other, MagicMock) and other.type == str(self.expected_type)


class GeneratorMock:
    def __init__(self, data):
        self.data = data
        self.time_extracted = datetime.datetime.utcnow().replace(tzinfo=pytz.UTC)
        self.endpoint_bookmark_field = "last_modified_date"
        self.endpoint_sorting_criteria = {'field': 'id'}
        self.client = "mock_client"
        self.state = "mock_state"
        self.config = "mock_config"

    def __iter__(self):
        return self

    def __next__(self):
        if not self.data:
            raise StopIteration
        return self.data.pop(0)

    def write_bookmark(self):
        pass


class ClientMock:
    def __init__(self, page_size=100):
        self.page_size = page_size


class MultithreadedOffsetGeneratorFake(MultithreadedOffsetGenerator):
    def __init__(self, stream_name='test_stream', client=ClientMock(), config=Mock(), state='', sub_type=''):
        self.stream_name = stream_name
        self.client = client
        self.config = config
        self.state = state
        self.sub_type = sub_type
        super().__init__(stream_name, client, config, state, sub_type)


class MultithreadedBookmarkGeneratorFake(MultithreadedBookmarkGenerator):
    def __init__(self, stream_name='test_stream', client=ClientMock(), config=Mock(), state='', sub_type=''):
        self.stream_name = stream_name
        self.client = client
        self.config = config
        self.state = state
        self.sub_type = sub_type
        super().__init__(stream_name, client, config, state, sub_type)


class MultithreadedBookmarkDayByDayGeneratorFake(MultithreadedBookmarkDayByDayGenerator):
    def __init__(self, stream_name='test_stream', client=ClientMock(), config=Mock(), state='', sub_type=''):
        self.stream_name = stream_name
        self.client = client
        self.config = config
        self.state = state
        self.sub_type = sub_type
        super().__init__(stream_name, client, config, state, sub_type)



from mock import MagicMock

from tap_mambu.tap_generators.child_generator import ChildGenerator
from ..constants import config_json
from tap_mambu.helpers.generator_processor_pairs import get_generator_processor_for_stream


def setup_generator_base_test(stream_name, client_mock=None):
    if client_mock is None:
        client_mock = MagicMock()
        client_mock.page_size = int(config_json.get("page_size", 500))
        client_mock.request = MagicMock()
    generator_classes, _ = get_generator_processor_for_stream(stream_name)

    generators = list()
    for generator_class in generator_classes:
        generator = generator_class(stream_name=stream_name,
                                    client=client_mock,
                                    config=config_json,
                                    state={"currently_syncing": stream_name},
                                    sub_type="self",
                                    **({"parent_id": "TEST"} if issubclass(generator_class, ChildGenerator) else {}))
        generators.append(generator)
    return generators



import re
import singer
from singer import write_state, Transformer, metadata


def get_bookmark(state, stream, sub_type, default):
    if (state is None) or ('bookmarks' not in state):
        return default

    if sub_type == 'self':
        return state.get('bookmarks', {}).get(stream, default)
    return state.get('bookmarks', {}).get(stream, {}).get(sub_type, default)


def transform_datetime(this_dttm):
    with Transformer() as transformer:
        # noinspection PyProtectedMember
        # pylint: disable=W0212
        new_dttm = transformer._transform_datetime(this_dttm)
    return new_dttm


def write_bookmark(state, stream, sub_type, value):
    if 'bookmarks' not in state:
        state['bookmarks'] = {}
    if stream not in state['bookmarks']:
        state['bookmarks'][stream] = {}
    if sub_type == 'self':
        state['bookmarks'][stream] = value
    else:
        if sub_type not in state['bookmarks'][stream]:
            state['bookmarks'][stream][sub_type] = {}
        state['bookmarks'][stream][sub_type] = value
    write_state(state)


# Convert camelCase to snake_case
def convert(name):
    reg_sub = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', reg_sub).lower()


# Convert keys in json array
def convert_array(arr):
    new_arr = []
    for i in arr:
        if isinstance(i, list):
            new_arr.append(convert_array(i))
        elif isinstance(i, dict):
            new_arr.append(convert_json(i))
        else:
            new_arr.append(i)
    return new_arr


# Convert keys in json
def convert_json(this_json):
    out = {}
    for key in this_json:
        new_key = convert(key)
        if isinstance(this_json[key], dict):
            out[new_key] = convert_json(this_json[key])
        elif isinstance(this_json[key], list):
            out[new_key] = convert_array(this_json[key])
        else:
            out[new_key] = this_json[key]
    return out


def remove_custom_nodes(this_json):
    if not isinstance(this_json, (dict, list)):
        return this_json
    if isinstance(this_json, list):
        return [remove_custom_nodes(vv) for vv in this_json]
    return {kk: remove_custom_nodes(vv) for kk, vv in this_json.items()
            if not kk[:1] == '_'}


def add_custom_field(key, record, custom_field_sets):
    for cf_key, cf_value in record.items():
        field = {
            'field_set_id': key,
            'id': cf_key,
            'value': cf_value,
        }
        custom_field_sets.append(field)


# Convert custom fields and sets
# Generalize/Abstract custom fields to key/value pairs
def convert_custom_fields(this_json):
    for record in this_json:
        cust_field_sets = []
        for key, value in record.items():
            if key.startswith('_'):
                if isinstance(value, dict):
                    add_custom_field(key, value, cust_field_sets)
                elif isinstance(value, list):
                    for element in value:
                        add_custom_field(key, element, cust_field_sets)
        record['custom_fields'] = cust_field_sets
    return this_json


# Run all transforms: denests _embedded, removes _embedded/_links, and
#  convert camelCase to snake_case for fieldname keys.
def transform_json(this_json, path):
    new_json = remove_custom_nodes(convert_custom_fields(this_json))
    out = dict()
    out[path] = new_json
    transformed_json = convert_json(out)
    return transformed_json[path]


# def transform_activities(this_json):
#     for record in this_json:
#         for key, value in record['activity'].items():
#             record[key] = value
#         del record['activity']
#     return this_json


# Review catalog and make a list of selected streams
def get_selected_streams(catalog):
    selected_streams = set()
    for stream in catalog.streams:
        mdata = metadata.to_map(stream.metadata)
        root_metadata = mdata.get(())
        if root_metadata and root_metadata.get('selected') is True:
            selected_streams.add(stream.tap_stream_id)
    return list(selected_streams)


# Currently syncing sets the stream currently being delivered in the state.
# If the integration is interrupted, this state property is used to identify
#  the starting point to continue from.
# Reference: https://github.com/singer-io/singer-python/blob/master/singer/bookmarks.py#L41-L46
def update_currently_syncing(state, stream_name):
    if (stream_name is None) and ('currently_syncing' in state):
        del state['currently_syncing']
    else:
        singer.set_currently_syncing(state, stream_name)
    singer.write_state(state)


# Review last_stream (last currently syncing stream), if any,
#  and continue where it left off in the selected streams.
# Or begin from the beginning, if no last_stream, and sync
#  all selected steams.
# Returns should_sync_stream (true/false) and last_stream.
def should_sync_stream(selected_streams, last_stream, stream_name):
    if last_stream == stream_name or last_stream is None:
        if last_stream is not None:
            last_stream = None
        if stream_name in selected_streams:
            return True, last_stream
    return False, last_stream



import requests
import pytest
import mock
from mock import Mock

from tap_mambu import MambuClient
from tap_mambu.helpers.client import MambuError, MambuNoCredInConfig, MambuNoSubdomainInConfig, \
    MambuNoAuditApikeyInConfig, Server5xxError, raise_for_error, ERROR_CODE_EXCEPTION_MAPPING
from .constants import config_json


@mock.patch("tap_mambu.helpers.client.MambuClient.check_access")
def test_client_entry_check_access(mock_check_access):
    client = MambuClient(username=config_json.get('username'),
                         password=config_json.get('password'),
                         subdomain=config_json['subdomain'],
                         page_size=int(config_json.get('page_size', 500)),
                         user_agent=config_json['user_agent'],
                         apikey='',
                         apikey_audit='')

    mock_check_access.assert_not_called()
    with client:
        mock_check_access.assert_called_once()


@mock.patch("tap_mambu.helpers.client.requests.Session.get")
def test_client_check_access_api_key(mock_requests_session_get):
    response = Mock()
    response.status_code = 200
    mock_requests_session_get.return_value = response

    client = MambuClient(username=config_json.get('username'),
                         password=None,
                         subdomain=config_json['subdomain'],
                         page_size=int(config_json.get('page_size', 500)),
                         user_agent='User_Agent_Test',
                         apikey='apikey_test',
                         apikey_audit='')
    base_url = "https://unit.test.mambu.com/api"
    assert client.check_access()
    assert client.base_url == base_url
    assert client.page_size == 500

    mock_requests_session_get.assert_called_once_with(url=f'{base_url}/settings/organization',
                                                      headers={'Accept': 'application/vnd.mambu.v1+json',
                                                               'User-Agent': 'MambuTap-User_Agent_Test'})


@mock.patch("tap_mambu.helpers.client.requests.Session.get")
def test_client_check_access_user_pass(mock_requests_session_get):
    response = Mock()
    response.status_code = 200
    mock_requests_session_get.return_value = response

    client = MambuClient(username=config_json.get('username'),
                         password=config_json.get('password'),
                         subdomain='test_subdomain',
                         page_size=100,
                         user_agent='',
                         apikey=None,
                         apikey_audit='')

    base_url = 'https://test_subdomain.mambu.com/api'
    assert client.check_access()
    assert client.base_url == base_url
    assert client.page_size == 100

    mock_requests_session_get.assert_called_once_with(url=f'{base_url}/settings/organization',
                                                      headers={'Accept': 'application/vnd.mambu.v1+json',
                                                               'User-Agent': 'MambuTap'})


@mock.patch("tap_mambu.helpers.client.raise_for_error")
@mock.patch("tap_mambu.helpers.client.requests.Session.get")
def test_client_check_status_code_fail(mock_requests_session_get, mock_raise_for_error):
    client = MambuClient(username=config_json.get('username'),
                         password=config_json.get('password'),
                         subdomain=config_json['subdomain'],
                         page_size=int(config_json.get('page_size', 500)),
                         user_agent=config_json['user_agent'],
                         apikey=None,
                         apikey_audit='')
    for status_code in [*range(0, 200), *range(201, 600)]:
        response = Mock()
        response.status_code = status_code
        mock_requests_session_get.return_value = response

        client.check_access()
        mock_raise_for_error.assert_called_with(response)


def test_client_check_access_no_auth_cred():
    client = MambuClient(username=None,
                         password=None,
                         subdomain=config_json['subdomain'],
                         page_size=int(config_json.get('page_size', 500)),
                         user_agent=config_json['user_agent'],
                         apikey=None,
                         apikey_audit='')

    with pytest.raises(MambuNoCredInConfig):
        client.check_access()


def test_client_check_access_no_subdomain():
    client = MambuClient(username=config_json.get('username'),
                         password=config_json.get('password'),
                         subdomain=None,
                         page_size=int(config_json.get('page_size', 500)),
                         user_agent=config_json['user_agent'],
                         apikey='',
                         apikey_audit='')

    with pytest.raises(MambuNoSubdomainInConfig):
        client.check_access()


@mock.patch("tap_mambu.helpers.client.metrics.http_request_timer")
@mock.patch("tap_mambu.helpers.client.requests.Session.request")
@mock.patch("tap_mambu.helpers.client.MambuClient.check_access")
def test_client_request_get_flow(mock_check_access, mock_requests_session_request, mock_metrics_http_request_timer):
    response = Mock()
    response.status_code = 200
    response.json.return_value = {'test_record': {'data1': 1, 'data2': 2}}
    mock_requests_session_request.return_value = response

    client = MambuClient(username=config_json.get('username'),
                         password=config_json.get('password'),
                         subdomain='test_subdomain',
                         page_size=int(config_json.get('page_size', 500)),
                         user_agent='test_user_agent',
                         apikey='',
                         apikey_audit='test_apikey_audit')

    req_response = client.request(method='GET', path='test_path', apikey_type='audit')
    mock_metrics_http_request_timer.assert_called_once_with(None)
    mock_requests_session_request.assert_called_once_with(method='GET',
                                                          url='https://test_subdomain.mambu.com/api/test_path',
                                                          json=None,
                                                          headers={'Accept': 'application/vnd.mambu.v2+json',
                                                                   'apikey': 'test_apikey_audit',
                                                                   'User-Agent': 'MambuTap-test_user_agent'})
    assert req_response == {'test_record': {'data1': 1, 'data2': 2}}

    client.request(method='POST', path='test_path', apikey_type='audit')
    mock_requests_session_request.assert_called_with(method='POST',
                                                     url='https://test_subdomain.mambu.com/api/test_path',
                                                     json=None,
                                                     headers={'Accept': 'application/vnd.mambu.v2+json',
                                                              'apikey': 'test_apikey_audit',
                                                              'User-Agent': 'MambuTap-test_user_agent',
                                                              'Content-Type': 'application/json'})
    mock_metrics_http_request_timer.assert_called_with(None)
    mock_check_access.assert_called_once()


@mock.patch("tap_mambu.helpers.client.metrics.http_request_timer")
@mock.patch("tap_mambu.helpers.client.requests.Session.request")
@mock.patch("tap_mambu.helpers.client.MambuClient.check_access")
def test_client_request_post_flow(mock_check_access, mock_requests_session_request, mock_metrics_http_request_timer):
    response = Mock()
    response.status_code = 200
    response.json.return_value = {}
    mock_requests_session_request.return_value = response

    client = MambuClient(username=config_json.get('username'),
                         password=config_json.get('password'),
                         subdomain=config_json['subdomain'],
                         page_size=int(config_json.get('page_size', 500)),
                         user_agent='',
                         apikey=None,
                         apikey_audit='')

    client.request(method='POST',
                   url='www.api.test_url',
                   json={'test': 'filter'},
                   version='v_test',
                   headers={'Accept': 'test_accept_header',
                            'User-Agent': 'test_user_agent',
                            'Content-Type': 'test_content_type'},
                   endpoint='test_endpoint_first')
    mock_metrics_http_request_timer.assert_called_once_with('test_endpoint_first')
    mock_requests_session_request.assert_called_once_with(method='POST',
                                                          url='www.api.test_url',
                                                          json={'test': 'filter'},
                                                          headers={'Accept': 'test_accept_header',
                                                                   'User-Agent': 'MambuTap',
                                                                   'Content-Type': 'application/json'})

    client.request(method='GET',
                   url='www.api.test_url',
                   json={'test': 'filter'},
                   version='v_test',
                   headers={'Accept': 'test_accept_header',
                            'User-Agent': 'test_user_agent',
                            'Content-Type': 'test_content_type'},
                   endpoint='test_endpoint_second')
    mock_requests_session_request.assert_called_with(method='GET',
                                                     url='www.api.test_url',
                                                     json={'test': 'filter'},
                                                     headers={'Accept': 'test_accept_header',
                                                              'User-Agent': 'MambuTap',
                                                              'Content-Type': 'test_content_type'})
    mock_metrics_http_request_timer.assert_called_with('test_endpoint_second')
    mock_check_access.assert_called_once()


@mock.patch("tap_mambu.helpers.client.MambuClient.check_access")
def test_client_request_no_audit_apikey(mock_check_access):
    client = MambuClient(username=config_json.get('username'),
                         password=config_json.get('password'),
                         subdomain=config_json['subdomain'],
                         page_size=int(config_json.get('page_size', 500)),
                         user_agent=config_json['user_agent'],
                         apikey='',
                         apikey_audit=None)

    with pytest.raises(MambuNoAuditApikeyInConfig):
        client.request(method='GET', apikey_type='audit')


@mock.patch("tap_mambu.helpers.client.backoff._decorator._sync.time.sleep")
@mock.patch("tap_mambu.helpers.client.metrics.http_request_timer")
@mock.patch("tap_mambu.helpers.client.requests.Session.request")
@mock.patch("tap_mambu.helpers.client.MambuClient.check_access")
def test_client_request_500_error(mock_check_access,
                                  mock_requests_session_request,
                                  mock_metrics_http_request_timer,
                                  mock_backoff_on_exception, ):
    client = MambuClient(username=config_json.get('username'),
                         password=config_json.get('password'),
                         subdomain=config_json['subdomain'],
                         page_size=int(config_json.get('page_size', 500)),
                         user_agent=config_json['user_agent'],
                         apikey='',
                         apikey_audit='')

    for status_code in range(500, 600):
        response = Mock()
        response.status_code = status_code
        mock_requests_session_request.return_value = response

        with pytest.raises(Server5xxError):
            client.request(method='GET')


@mock.patch("tap_mambu.helpers.client.raise_for_error")
@mock.patch("tap_mambu.helpers.client.metrics.http_request_timer")
@mock.patch("tap_mambu.helpers.client.requests.Session.request")
@mock.patch("tap_mambu.helpers.client.MambuClient.check_access")
def test_client_request_raise_for_error(mock_check_access,
                                        mock_requests_session_request,
                                        mock_metrics_http_request_timer,
                                        mock_raise_for_error):
    client = MambuClient(username=config_json.get('username'),
                         password=config_json.get('password'),
                         subdomain=config_json['subdomain'],
                         page_size=int(config_json.get('page_size', 500)),
                         user_agent=config_json['user_agent'],
                         apikey='',
                         apikey_audit='')

    for status_code in [*range(0, 200), *range(201, 500)]:
        response = Mock()
        response.status_code = status_code
        mock_requests_session_request.return_value = response

        client.request(method='GET')
        mock_raise_for_error.assert_called_with(response)


@mock.patch("tap_mambu.helpers.client.requests.models.Response.raise_for_status")
def test_raise_for_error_call(mock_requests_raise_for_status):
    response = Mock()
    response.raise_for_status = mock_requests_raise_for_status

    raise_for_error(response)

    # tested if the raise_for_status function was called
    mock_requests_raise_for_status.assert_called_once()


@mock.patch("tap_mambu.helpers.client.requests.models.Response.raise_for_status")
def test_raise_for_error_no_content(mock_requests_raise_for_status):
    mock_requests_raise_for_status.side_effect = Mock(side_effect=requests.HTTPError('test'))
    response = Mock()
    response.content = []
    response.raise_for_status = mock_requests_raise_for_status

    func_response = raise_for_error(response)

    # test case: when response content is empty -> the function should just return and do nothing
    assert func_response is None
    response.json.assert_not_called()

    # test case: when something from the try/except HTTPError block throws ValueError, TypeError ->
    # it should raise MambuError
    response.content = None
    with pytest.raises(MambuError):
        raise_for_error(response)
    response.json.assert_not_called()

    # test case: when the response content isn't empty but it also doesn't contain "error" or "errorCode" ->
    # it should raise MambuError
    response.content = {'error': 404,
                        'status': 404,
                        'message': 'not found'}
    response.json.return_value = {}
    with pytest.raises(MambuError):
        raise_for_error(response)
    response.json.assert_called_once()


@mock.patch("tap_mambu.helpers.client.requests.models.Response.raise_for_status")
def test_raise_for_error_known_error_codes(mock_requests_raise_for_status):
    mock_requests_raise_for_status.side_effect = Mock(side_effect=requests.HTTPError('test'))

    for status_code in range(300, 501):
        content = {'error': status_code,
                   'status': status_code,
                   'message': 'test'}
        response = Mock()
        response.content = content
        response.json.return_value = content
        response.raise_for_status = mock_requests_raise_for_status

        # test if the exceptions are handled correctly, meaning ->
        # - throw the custom exceptions for the defined ones
        # - throw MambuError for others
        if status_code not in ERROR_CODE_EXCEPTION_MAPPING:
            with pytest.raises(MambuError):
                raise_for_error(response)
        else:
            with pytest.raises(ERROR_CODE_EXCEPTION_MAPPING[status_code]):
                raise_for_error(response)
        response.json.assert_called_once()



from .processor import TapProcessor, LOGGER
from ..helpers import get_selected_streams
from ..helpers.multithreaded_requests import MultithreadedRequestsPool


class MultithreadedParentProcessor(TapProcessor):
    def _process_child_records_multithreaded(self, record):
        from ..sync import sync_endpoint

        super(MultithreadedParentProcessor, self)._process_child_records_multithreaded(record)
        for child_stream_name in self.endpoint_child_streams:
            if child_stream_name in get_selected_streams(self.catalog):
                parent_id = record[self.endpoint_id_field]
                LOGGER.info(f'(processor) Syncing: {child_stream_name}, '
                            f'parent_stream: {self.stream_name}, '
                            f'parent_id: {parent_id}')

                future = MultithreadedRequestsPool.queue_function(
                    sync_endpoint,
                    client=self.client,
                    catalog=self.catalog,
                    state=self.state,
                    stream_name=child_stream_name,
                    sub_type=self.sub_type,
                    config=self.config,
                    parent_id=parent_id)
                self.futures.append(future)



from .generator import TapGenerator


class CustomFieldSetsGenerator(TapGenerator):
    def _init_endpoint_config(self):
        super(CustomFieldSetsGenerator, self)._init_endpoint_config()
        self.endpoint_path = "customfieldsets"
        self.endpoint_api_method = "GET"



import datetime
import json
import time
import backoff

from copy import deepcopy
from singer import get_logger

from .multithreaded_offset_generator import MultithreadedOffsetGenerator
from ..helpers import transform_json, convert, transform_datetime
from ..helpers.multithreaded_requests import MultithreadedRequestsPool

LOGGER = get_logger()


class MultithreadedBookmarkGenerator(MultithreadedOffsetGenerator):
    def _init_endpoint_config(self):
        super(MultithreadedBookmarkGenerator, self)._init_endpoint_config()
        self.endpoint_intermediary_bookmark_value = None
        self.endpoint_intermediary_bookmark_offset = 0

    def prepare_batch_params(self):
        self.offset = self.endpoint_intermediary_bookmark_offset
        # This function should be extended upon in case you wish to change the date to a new one,
        # in order to paginate through the data using date, resetting offset to 0, or the specific number of records
        # the next batch needs to skip if there are multiple records with the same date.

    def fetch_batch_continuously(self):
        first_run = True
        while not self.end_of_file:
            while len(self.buffer) > self.batch_limit:
                time.sleep(0.1)
            if not first_run:
                self.prepare_batch_params()
            if not self._all_fetch_batch_steps():
                self.end_of_file = True
            first_run = False

    def queue_batches(self):
        # prepare batches (with self.limit for each of them until we reach batch_limit)
        futures = list()
        original_offset = self.offset
        for offset in range(0, self.batch_limit, self.artificial_limit):
            self.offset = original_offset + offset
            self.prepare_batch()
            # send batches to multithreaded_requests_pool
            futures.append(MultithreadedRequestsPool.queue_request(self.client, self.stream_name,
                                                                   self.endpoint_path, self.endpoint_api_method,
                                                                   self.endpoint_api_version,
                                                                   self.endpoint_api_key_type,
                                                                   deepcopy(self.endpoint_body),
                                                                   deepcopy(self.params)))
        return futures

    def collect_batches(self, futures):
        # wait for responses, and check them for errors
        final_buffer = set()
        stop_iteration = False
        for future in futures:
            while not future.done():
                time.sleep(0.1)
            result = future.result()
            transformed_batch = self.transform_batch(transform_json(result, self.stream_name))
            temp_buffer = set(transformed_batch)

            if not final_buffer:
                final_buffer = final_buffer | temp_buffer
                continue

            if not temp_buffer:  # We finished the data to extract, time to stop
                self.stop_all_request_threads(futures)
                stop_iteration = True
                break

            final_buffer = self.error_check_and_fix(final_buffer, temp_buffer, futures)

            if stop_iteration:
                break
        return final_buffer, stop_iteration

    def preprocess_record(self, raw_record):
        record = super().preprocess_record(raw_record)
        # increment bookmark
        record_bookmark_value = record.get(convert(self.endpoint_bookmark_field))
        if record_bookmark_value is not None:
            self.set_intermediary_bookmark(transform_datetime(record_bookmark_value))
        return record

    def preprocess_batches(self, final_buffer):
        super().preprocess_batches(final_buffer)
        self.last_batch_size = len(final_buffer)

    def set_intermediary_bookmark(self, record_bookmark_value):
        if self.endpoint_intermediary_bookmark_value is None or \
                self.compare_bookmark_values(record_bookmark_value,
                                             self.endpoint_intermediary_bookmark_value):
            self.endpoint_intermediary_bookmark_value = record_bookmark_value
            self.endpoint_intermediary_bookmark_offset = 1
            return

        if record_bookmark_value == self.endpoint_intermediary_bookmark_value:
            self.endpoint_intermediary_bookmark_offset += 1
            return

    def compare_bookmark_values(self, a, b):
        return a > b


class MultithreadedBookmarkDayByDayGenerator(MultithreadedBookmarkGenerator):
    def set_intermediary_bookmark(self, record_bookmark_value):
        super().set_intermediary_bookmark(record_bookmark_value[:10])



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(name='tap-mambu',
      version='3.1.2',
      description='Singer.io tap for extracting data from the Mambu 2.0 API',
      author='jeff.huth@bytecode.io',
      classifiers=['Programming Language :: Python :: 3 :: Only'],
      py_modules=['tap_mambu'],
      python_requires='>=3.9',
      install_requires=[
          'backoff==1.8.0',
          'requests==2.23.0',
          'singer-python==5.12.2',
          'pytz==2022.1'
      ],
      extras_require={
          'dev': [
              'ipdb',
              'pylint==2.5.3',
              'pytest==6.2.4'
          ],
          'mambu-tests': [
              'coverage==6.3.1',
              'pylint==2.12.2',
              'pytest==6.2.5',
              'mock==4.0.3'
          ],
          'mambu-performance': [
              'matplotlib==3.5.1'
          ]
      },
      entry_points='''
          [console_scripts]
          tap-mambu=tap_mambu:main
      ''',
      packages=find_packages(),
      package_data={
          'tap_mambu': [
              'helpers/schemas/*.json'
          ]
      })



from . import setup_generator_base_test


def test_groups_generator():
    generators = setup_generator_base_test("groups")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.endpoint_path == "groups:search"
    assert generator.endpoint_bookmark_field == "lastModifiedDate"
    assert generator.endpoint_sorting_criteria == {
        "field": "lastModifiedDate",
        "order": "ASC"
    }
    assert generator.endpoint_filter_criteria == [
        {
            "field": "lastModifiedDate",
            "operator": "AFTER",
            "value": '2021-06-01T00:00:00.000000Z'
        }
    ]



import os
import json
from singer import metadata

# Reference:
# https://github.com/singer-io/getting-started/blob/master/docs/DISCOVERY_MODE.md#Metadata
STREAMS = {
    'branches': {
        'key_properties': ['id'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['last_modified_date']
    },
    'cards': {
        'key_properties': ['deposit_id', 'reference_token'],
        'replication_method': 'FULL_TABLE'
    },
    'communications': {
        'key_properties': ['encoded_key'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['creation_date']
    },
    'centres': {
        'key_properties': ['id'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['last_modified_date']
    },
    'clients': {
        'key_properties': ['id'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['last_modified_date']
    },
    'credit_arrangements': {
        'key_properties': ['id'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['last_modified_date']
    },
    'custom_field_sets': {
        'key_properties': ['id'],
        'replication_method': 'FULL_TABLE'
    },
    'deposit_accounts': {
        'key_properties': ['id'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['last_modified_date']
    },
    'deposit_products': {
        'key_properties': ['id'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['last_modified_date']
    },
    'deposit_transactions': {
        'key_properties': ['encoded_key'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['creation_date']
    },
    'groups': {
        'key_properties': ['id'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['last_modified_date']
    },
    'loan_accounts': {
        'key_properties': ['id'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['last_modified_date', 'last_account_appraisal_date']
    },
    'loan_repayments': {
        'key_properties': ['encoded_key'],
        'replication_method': 'FULL_TABLE'
    },
    'loan_products': {
        'key_properties': ['id'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['last_modified_date']
    },
    'loan_transactions': {
        'key_properties': ['encoded_key'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['creation_date']
    },
    'tasks': {
        'key_properties': ['id'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['last_modified_date']
    },
    'users': {
        'key_properties': ['id'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['last_modified_date']
    },
    'gl_accounts': {
        'key_properties': ['gl_code'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['last_modified_date']
    },
    'gl_journal_entries': {
        'key_properties': ['entry_id'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['creation_date']
    },
    'activities': {
        'key_properties': ['encoded_key'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['timestamp']
    },
    'index_rate_sources': {
        'key_properties': ['encoded_key'],
        'replication_method': 'FULL_TABLE'
    },
    'installments': {
        'key_properties': ['encoded_key'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['last_paid_date']
    },
    'audit_trail': {
        'key_properties': [],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['occurred_at']
    },
    'interest_accrual_breakdown': {
        'key_properties': ['entry_id'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['creation_date']
    },
}


def get_abs_path(path):
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)


def get_schemas():
    schemas = {}
    field_metadata = {}

    for stream_name, stream_metadata in STREAMS.items():
        schema_path = get_abs_path('schemas/{}.json'.format(stream_name))
        with open(schema_path) as file:
            schema = json.load(file)
        schemas[stream_name] = schema
        mdata = metadata.new()

        # Documentation:
        # https://github.com/singer-io/getting-started/blob/master/docs/DISCOVERY_MODE.md#singer-python-helper-functions
        # Reference:
        # https://github.com/singer-io/singer-python/blob/master/singer/metadata.py#L25-L44
        mdata = metadata.get_standard_metadata(
            schema=schema,
            key_properties=stream_metadata.get('key_properties', None),
            valid_replication_keys=stream_metadata.get('replication_keys', None),
            replication_method=stream_metadata.get('replication_method', None)
        )
        field_metadata[stream_name] = mdata

    return schemas, field_metadata



from . import setup_generator_base_test


def test_deposit_products_generator_endpoint_config_init():
    generators = setup_generator_base_test("deposit_products")

    assert 1 == len(generators)

    generator = generators[0]

    assert generator.endpoint_path == "depositproducts"
    assert generator.endpoint_api_method == "GET"
    assert generator.endpoint_bookmark_field == "lastModifiedDate"
